'use strict';

var core = require('@math-blocks/core');
var Editor = require('@math-blocks/editor');

function _interopNamespaceDefault(e) {
    var n = Object.create(null);
    if (e) {
        Object.keys(e).forEach(function (k) {
            if (k !== 'default') {
                var d = Object.getOwnPropertyDescriptor(e, k);
                Object.defineProperty(n, k, d.get ? d : {
                    enumerable: true,
                    get: function () { return e[k]; }
                });
            }
        });
    }
    n.default = e;
    return Object.freeze(n);
}

var Editor__namespace = /*#__PURE__*/_interopNamespaceDefault(Editor);

exports.MathStyle = void 0;
(function (MathStyle) {
    MathStyle["Display"] = "D";
    MathStyle["Text"] = "T";
    MathStyle["Script"] = "S";
    MathStyle["ScriptScript"] = "SS";
})(exports.MathStyle || (exports.MathStyle = {}));
exports.RenderMode = void 0;
(function (RenderMode) {
    RenderMode["Static"] = "static";
    RenderMode["Dynamic"] = "dynamic";
})(exports.RenderMode || (exports.RenderMode = {}));
exports.RadicalDegreeAlgorithm = void 0;
(function (RadicalDegreeAlgorithm) {
    RadicalDegreeAlgorithm[RadicalDegreeAlgorithm["OpenType"] = 0] = "OpenType";
    RadicalDegreeAlgorithm[RadicalDegreeAlgorithm["MathML"] = 1] = "MathML";
})(exports.RadicalDegreeAlgorithm || (exports.RadicalDegreeAlgorithm = {}));

const makeHBox = (dim, content, context) => {
    return {
        type: 'HBox',
        ...dim,
        shift: 0,
        content,
        fontSize: fontSizeForContext(context),
        style: {},
    };
};
const rebox = (box, before, after) => {
    if (box.content.type === 'static') {
        return {
            ...box,
            width: before.size + box.width + after.size,
            content: {
                ...box.content,
                nodes: [before, ...box.content.nodes, after],
            },
        };
    }
    else {
        return {
            ...box,
            width: before.size + box.width + after.size,
            content: {
                ...box.content,
                left: [before, ...box.content.left],
                right: [...box.content.right, after],
            },
        };
    }
};
const makeKern = (size, flag) => ({
    type: 'Kern',
    size,
    flag,
    style: {},
});
const makeHRule = (thickness, width) => ({
    type: 'HRule',
    thickness,
    width,
    style: {},
});
const makeGlyph = (char, glyphID, context, isDelimiter = false) => {
    return {
        type: 'Glyph',
        char,
        glyphID,
        size: fontSizeForContext(context),
        fontData: context.fontData,
        style: {},
        isDelimiter,
    };
};
const getCharAdvance = (glyph) => {
    const { font } = glyph.fontData;
    const metrics = font.getGlyphMetrics(glyph.glyphID);
    const unitsPerEm = font.head.unitsPerEm;
    if (!metrics) {
        throw new Error(`metrics do not exist for "${glyph.char}"`);
    }
    return (metrics.advance * glyph.size) / unitsPerEm;
};
const getCharHeight = (glyph) => {
    const { font } = glyph.fontData;
    const metrics = font.getGlyphMetrics(glyph.glyphID);
    const unitsPerEm = font.head.unitsPerEm;
    if (!metrics) {
        throw new Error(`metrics do not exist for "${glyph.char}"`);
    }
    return (metrics.bearingY * glyph.size) / unitsPerEm;
};
const getCharDepth = (glyph) => {
    const { font } = glyph.fontData;
    const metrics = font.getGlyphMetrics(glyph.glyphID);
    const unitsPerEm = font.head.unitsPerEm;
    if (!metrics) {
        throw new Error(`metrics do not exist for "${glyph.char}"`);
    }
    return ((metrics.height - metrics.bearingY) * glyph.size) / unitsPerEm;
};
const getWidth = (node) => {
    switch (node.type) {
        case 'HBox':
            return node.width;
        case 'VBox':
            return node.width;
        case 'Glyph':
            return getCharAdvance(node);
        case 'Kern':
            return node.size;
        case 'HRule':
            return node.width;
        default:
            throw new core.UnreachableCaseError(node);
    }
};
const getHeight = (node) => {
    switch (node.type) {
        case 'HBox':
            return node.height - node.shift;
        case 'VBox':
            return node.height - node.shift;
        case 'Glyph':
            return getCharHeight(node);
        case 'Kern':
            return 0;
        case 'HRule':
            return node.thickness / 2;
        default:
            throw new core.UnreachableCaseError(node);
    }
};
const getDepth = (node) => {
    switch (node.type) {
        case 'HBox':
            return node.depth + node.shift;
        case 'VBox':
            return node.depth + node.shift;
        case 'Glyph':
            return getCharDepth(node);
        case 'Kern':
            return 0;
        case 'HRule':
            return node.thickness / 2;
        default:
            throw new core.UnreachableCaseError(node);
    }
};
const vsize = (node) => {
    switch (node.type) {
        case 'HBox':
            return node.height + node.depth;
        case 'VBox':
            return node.height + node.depth;
        case 'Glyph':
            return getCharHeight(node) + getCharDepth(node);
        case 'Kern':
            return node.size;
        case 'HRule':
            return node.thickness;
        default:
            throw new core.UnreachableCaseError(node);
    }
};
const add = (a, b) => a + b;
const zero = 0;
const sum = (values) => values.reduce(add, zero);
const max = (values) => Math.max(...values);
const hlistWidth = (nodes) => sum(nodes.map(getWidth));
const hlistHeight = (nodes) => max(nodes.map(getHeight));
const hlistDepth = (nodes) => max(nodes.map(getDepth));
const vlistVsize = (nodes) => sum(nodes.map(vsize));
const makeStaticHBox = (nodes, context) => {
    const dim = {
        width: hlistWidth(nodes),
        height: hlistHeight(nodes),
        depth: hlistDepth(nodes),
    };
    const content = {
        type: 'static',
        nodes,
    };
    return makeHBox(dim, content, context);
};
const makeCursorHBox = (left, right, context) => {
    const dim = {
        width: hlistWidth(left) + hlistWidth(right),
        height: Math.max(hlistHeight(left), hlistHeight(right)),
        depth: Math.max(hlistDepth(left), hlistDepth(right)),
    };
    const content = {
        type: 'cursor',
        left,
        right,
    };
    return makeHBox(dim, content, context);
};
const makeSelectionHBox = (left, selection, right, context) => {
    const dim = {
        width: hlistWidth(left) + hlistWidth(selection) + hlistWidth(right),
        height: Math.max(hlistHeight(left), hlistHeight(selection), hlistHeight(right)),
        depth: Math.max(hlistDepth(left), hlistDepth(selection), hlistDepth(right)),
    };
    const content = {
        type: 'selection',
        left,
        selection,
        right,
    };
    return makeHBox(dim, content, context);
};
const makeVBox = (width, node, upList, dnList, context) => {
    const dim = {
        width,
        depth: dnList.length > 0 ? vlistVsize(dnList) + getDepth(node) : getDepth(node),
        height: upList.length > 0
            ? vlistVsize(upList) + getHeight(node)
            : getHeight(node),
    };
    const upListCopy = [...upList];
    // TODO: get rid of the need to reverse the uplist
    const nodeList = [...upListCopy.reverse(), node, ...dnList];
    return {
        type: 'VBox',
        ...dim,
        shift: 0,
        content: nodeList,
        fontSize: fontSizeForContext(context),
        style: {},
    };
};
const getConstantValue = (constant, context) => {
    return constant * fontSizeForContext(context);
};
// TODO: return the font size instead of the multiplier
const multiplierForContext = (context) => {
    const { constants } = context.fontData.font.math;
    const { mathStyle } = context;
    switch (mathStyle) {
        case exports.MathStyle.Display:
        case exports.MathStyle.Text:
            return 1.0;
        case exports.MathStyle.Script:
            return constants.scriptPercentScaleDown / 100;
        case exports.MathStyle.ScriptScript:
            return constants.scriptScriptPercentScaleDown / 100;
    }
};
const fontSizeForContext = (context) => {
    const multiplier = multiplierForContext(context);
    const { baseFontSize } = context;
    const fontSize = multiplier * baseFontSize;
    return fontSize;
};
// TODO: special case how we compute the delimiters for rows including "y" or
// other deep descenders so that we get the same font size as the rest of the
// glyphs on that row.
const getDelimiter = (char, box, thresholdOptions, context) => {
    const { font } = context.fontData;
    const glyphID = font.getGlyphID(char);
    const construction = font.math.variants.getVertGlyphConstruction(glyphID);
    if (!construction) {
        return glyphID;
    }
    const fontSize = fontSizeForContext(context);
    for (let i = 0; i < construction.mathGlyphVariantRecords.length; i++) {
        const record = construction.mathGlyphVariantRecords[i];
        const glyphMetrics = font.getGlyphMetrics(record.variantGlyph);
        const height = (glyphMetrics.bearingY * fontSize) / font.head.unitsPerEm;
        const depth = ((glyphMetrics.height - glyphMetrics.bearingY) * fontSize) /
            font.head.unitsPerEm;
        const compare = thresholdOptions.strict
            ? (a, b) => a > b
            : (a, b) => a >= b;
        // TODO: add an option to configure whether these inequalities are
        // strict or not.
        switch (thresholdOptions.value) {
            case 'both': {
                if (compare(height, box.height) && compare(depth, box.depth)) {
                    // HACK: this is to ensure that we're using the same size
                    // glyph as the row when it contains deep descenders like "y"
                    if (i === 1 && char !== '\u221a') {
                        return glyphID;
                    }
                    return record.variantGlyph;
                }
                break;
            }
            case 'sum': {
                if (compare(height + depth, box.height + box.depth)) {
                    // HACK: this is to ensure that we're using the same size
                    // glyph as the row when it contains deep descenders like "y"
                    if (i === 1 && char !== '\u221a') {
                        return glyphID;
                    }
                    return record.variantGlyph;
                }
                break;
            }
        }
    }
    if (construction.mathGlyphVariantRecords.length > 0) {
        // TODO: return a glyph assembly layout instead of the tallest delim
        return construction.mathGlyphVariantRecords[construction.mathGlyphVariantRecords.length - 1].variantGlyph;
    }
    return glyphID;
};
/**
 * Returns the smallest delimiter whose depth and height exceed the that of the
 * box passed in.  `thresholdOptions` controls how this comparison is made.
 *
 * @param {string} char
 * @param {Box} box layout encompassed by the delimiter
 * @param {ThresholdOptions} options
 * @param {Context} context
 * @returns {Glyph}
 */
const makeDelimiter = (char, box, thresholdOptions, context) => {
    const glyphID = getDelimiter(char, box, thresholdOptions, context);
    return makeGlyph(char, glyphID, context, true);
};

const processHRule = (hrule, loc) => {
    const advance = getWidth(hrule);
    return {
        type: 'line',
        x1: loc.x,
        y1: loc.y,
        x2: loc.x + advance,
        y2: loc.y,
        thickness: hrule.thickness,
        style: {
            stroke: hrule.style.color,
        },
        id: hrule.id,
    };
};
const processGlyph = (glyph, loc) => {
    return {
        type: 'char',
        x: loc.x,
        y: loc.y,
        width: getWidth(glyph),
        glyph: glyph,
        style: {
            fill: glyph.style.color,
        },
        id: glyph.id,
    };
};
const CURSOR_WIDTH = 1;
const processHBox = (box, loc, context) => {
    const pen = { x: 0, y: 0 };
    const selectionBoxes = [];
    const children = [];
    const hasSelection = !context.inSelection && box.content.type === 'selection';
    const { layer, fontData: { font }, } = context;
    const { fontSize } = box;
    const parenMetrics = font.getGlyphMetrics(font.getGlyphID(')'));
    // This assumes that parenMetrics.height < font.head.unitsPerEm
    const overshoot = (font.head.unitsPerEm - parenMetrics.height) / 2;
    const ascent = ((parenMetrics.bearingY + overshoot) * fontSize) / font.head.unitsPerEm;
    const debugStyle = {
        fill: 'none',
        stroke: 'red',
    };
    let cancelRegion = null;
    const sections = [];
    if (box.content.type === 'static') {
        sections.push(box.content.nodes);
    }
    else if (box.content.type === 'cursor') {
        sections.push(box.content.left);
        sections.push(box.content.right);
    }
    else {
        sections.push(box.content.left);
        sections.push(box.content.selection);
        sections.push(box.content.right);
    }
    sections.forEach((section, index) => {
        const isSelection = hasSelection && index === 1;
        // There should only be two sections max.  If there are two sections
        // then we should draw a cursor in between the two of them.
        if (index === 1 &&
            !hasSelection &&
            !context.inSelection &&
            context.showCursor &&
            layer === 'selection') {
            // Draw the cursor.
            children.push({
                type: 'rect',
                x: pen.x - CURSOR_WIDTH / 2,
                y: pen.y - ascent,
                width: CURSOR_WIDTH,
                height: fontSize,
                style: {},
                className: 'blink',
                // We give the cursor rect a unique key whenever we re-render
                // the scene-graph so that it's always visible when moving the
                // cursor.
                key: Math.random().toString(),
            });
        }
        section.forEach((node) => {
            // We've encountered a new cancel id or no cancel id.  Finalize the
            // current cancel region and reset for the next one.
            if (cancelRegion &&
                cancelRegion.id !== node.style.cancel &&
                layer === 'content') {
                children.push({
                    type: 'line',
                    x1: cancelRegion.xMin,
                    y1: cancelRegion.yMax,
                    x2: cancelRegion.xMax,
                    y2: cancelRegion.yMin,
                    thickness: 5,
                    style: {},
                });
                cancelRegion = null;
            }
            // If the node has a cancel id, continue the current cancel region
            // or start a new one if there isn't one.
            if (typeof node.style.cancel === 'number' && layer == 'content') {
                const yMin = -Math.max(getHeight(node), ascent);
                const height = Math.max(getHeight(node) + getDepth(node), fontSize);
                const yMax = yMin + height;
                if (!cancelRegion) {
                    cancelRegion = {
                        id: node.style.cancel,
                        xMin: pen.x,
                        xMax: pen.x,
                        yMin,
                        yMax,
                    };
                }
                else {
                    cancelRegion.yMin = Math.min(cancelRegion.yMin, yMin);
                    cancelRegion.yMax = Math.max(cancelRegion.yMax, yMax);
                }
            }
            if (isSelection && layer === 'selection') {
                const yMin = -Math.max(getHeight(node), ascent);
                const height = Math.max(getHeight(node) + getDepth(node), fontSize);
                // Draw the cursor.
                selectionBoxes.push({
                    type: 'rect',
                    x: pen.x,
                    y: yMin,
                    width: getWidth(node),
                    height: height,
                    style: {},
                });
            }
            const advance = getWidth(node);
            const height = getHeight(node);
            const depth = getDepth(node);
            switch (node.type) {
                case 'HBox':
                case 'VBox': {
                    const child = _processBox(node, { x: pen.x, y: pen.y + node.shift }, {
                        ...context,
                        inSelection: context.inSelection || hasSelection,
                    });
                    // We always have to include child groups regardless of the
                    // layer.  TODO: drop this once we flatten the scene graph.
                    children.push(child);
                    if (layer === 'debug') {
                        children.push({
                            type: 'rect',
                            id: node.id,
                            x: pen.x,
                            y: pen.y - height,
                            width: advance,
                            height: depth + height,
                            style: debugStyle,
                        });
                    }
                    break;
                }
                case 'HRule': {
                    const child = processHRule(node, pen);
                    if (layer === 'content') {
                        children.push(child);
                    }
                    break;
                }
                case 'Glyph': {
                    const child = processGlyph(node, pen);
                    if (layer === 'content') {
                        children.push(child);
                    }
                    if (layer === 'debug') {
                        children.push({
                            type: 'rect',
                            id: node.id,
                            x: pen.x,
                            y: pen.y - height,
                            width: advance,
                            height: depth + height,
                            style: debugStyle,
                        });
                    }
                    if (layer === 'hitboxes') {
                        // TODO: do a second pass on the hitboxes to expand them
                        // to their full height
                        children.push({
                            type: 'rect',
                            id: node.id,
                            x: pen.x,
                            y: pen.y - height,
                            width: advance,
                            height: depth + height,
                            style: debugStyle,
                        });
                    }
                    break;
                }
                case 'Kern':
                    if (node.flag) {
                        if (layer === 'hitboxes') {
                            // TODO: do a second pass on the hitboxes to expand
                            // them to their full height
                            children.push({
                                type: 'rect',
                                flag: node.flag,
                                x: pen.x,
                                y: pen.y - box.height,
                                width: node.size,
                                height: box.depth + box.height,
                                style: debugStyle,
                            });
                        }
                    }
                    // We don't need to include kerns in the output since we include
                    // the cursor or select rectangle in the scene graph.
                    break;
                default:
                    throw new core.UnreachableCaseError(node);
            }
            pen.x += advance;
            if (node.style.cancel && cancelRegion) {
                cancelRegion.xMax = pen.x;
            }
        });
    });
    if (cancelRegion !== null && layer === 'content') {
        const thickness = fontSize * font.math.constants.fractionRuleThickness;
        children.push({
            type: 'line',
            // @ts-expect-error: TypeScript doesn't understand forEach loops
            x1: cancelRegion.xMin,
            // @ts-expect-error: TypeScript doesn't understand forEach loops
            y1: cancelRegion.yMax,
            // @ts-expect-error: TypeScript doesn't understand forEach loops
            x2: cancelRegion.xMax,
            // @ts-expect-error: TypeScript doesn't understand forEach loops
            y2: cancelRegion.yMin,
            thickness: thickness,
            style: {},
        });
        cancelRegion = null;
    }
    // Draw the selection.
    if (layer === 'selection') {
        for (const selectionBox of selectionBoxes) {
            children.unshift({
                ...selectionBox,
                fill: 'Highlight',
                stroke: 'none',
            });
        }
    }
    return {
        type: 'group',
        x: loc.x,
        y: loc.y,
        bounds: box,
        children: children,
        style: layer === 'content'
            ? {
                fill: box.style.color,
                stroke: box.style.color,
            }
            : {},
        id: box.id,
    };
};
const processVBox = (box, loc, context) => {
    const pen = { x: 0, y: 0 };
    pen.y -= box.height;
    const children = [];
    const { layer } = context;
    const debugStyle = {
        fill: 'none',
        stroke: 'red',
    };
    box.content.forEach((node) => {
        const width = getWidth(node);
        const height = getHeight(node);
        const depth = getDepth(node);
        switch (node.type) {
            case 'HBox':
            case 'VBox': {
                // TODO: reconsider whether we should be taking the shift into
                // account when computing the height, maybe we can drop this
                // and simplify things.  The reason why we zero out the shift
                // here is that when we render a box inside of a vbox, the shift
                // is a horizontal shift as opposed to a vertical one.
                // I'm not sure we can do this properly since how the shift is
                // used depends on the parent box type.  We could pass that info
                // to the getHeight() function... we should probably do an audit
                // of all the callsites for getHeight()
                const height = getHeight({ ...node, shift: 0 });
                const depth = getDepth({ ...node, shift: 0 });
                pen.y += height;
                // TODO: see if we can get rid of this check in the future
                if (Number.isNaN(pen.y)) {
                    // eslint-disable-next-line no-debugger
                    debugger;
                }
                const child = _processBox(node, { x: pen.x + node.shift, y: pen.y }, context);
                // We always have to include child groups regardless of the
                // layer.  TODO: drop this once we flatten the scene graph.
                children.push(child);
                if (layer === 'debug') {
                    children.push({
                        type: 'rect',
                        id: node.id,
                        x: pen.x + node.shift,
                        y: pen.y - height,
                        width: width,
                        height: depth + height,
                        style: debugStyle,
                    });
                }
                pen.y += depth;
                break;
            }
            case 'HRule': {
                pen.y += height;
                const child = processHRule(node, pen);
                if (layer === 'content') {
                    children.push(child);
                }
                pen.y += depth;
                break;
            }
            case 'Glyph': {
                // Although there currently isn't anything that uses a glyph
                // in a vbox, we'll likely need it for accents.
                pen.y += height;
                const child = processGlyph(node, pen);
                if (layer === 'content') {
                    children.push(child);
                }
                if (layer === 'debug') {
                    children.push({
                        type: 'rect',
                        id: node.id,
                        x: pen.x,
                        y: pen.y,
                        width: width,
                        height: depth + height,
                        style: debugStyle,
                    });
                }
                if (layer === 'hitboxes') {
                    // TODO: do a second pass on the hitboxes to expand them
                    // to their full height
                    children.push({
                        type: 'rect',
                        id: node.id,
                        x: pen.x,
                        y: pen.y,
                        width: width,
                        height: depth + height,
                        style: debugStyle,
                    });
                }
                pen.y += depth;
                break;
            }
            case 'Kern':
                pen.y += node.size;
                break;
            default:
                throw new core.UnreachableCaseError(node);
        }
    });
    return {
        type: 'group',
        x: loc.x,
        y: loc.y,
        bounds: box,
        children: children,
        style: layer === 'content'
            ? {
                fill: box.style.color,
                stroke: box.style.color,
            }
            : {},
        id: box.id,
    };
};
const _processBox = (box, loc, context) => {
    switch (box.type) {
        case 'HBox':
            return processHBox(box, loc, context);
        case 'VBox':
            return processVBox(box, loc, context);
    }
};
const processBox = (box, fontData, options = {}) => {
    const loc = { x: 0, y: getHeight(box) };
    const context = {
        showCursor: !!options.showCursor,
        inSelection: false,
        fontData: fontData,
        layer: 'content',
    };
    const contentLayer = _processBox(box, loc, context);
    const selectionLayer = _processBox(box, loc, {
        ...context,
        layer: 'selection',
    });
    const debugLayer = _processBox(box, loc, { ...context, layer: 'debug' });
    const hitboxes = _processBox(box, loc, { ...context, layer: 'hitboxes' });
    const scene = {
        width: Math.ceil(contentLayer.bounds.width),
        height: Math.ceil(contentLayer.bounds.height + contentLayer.bounds.depth),
        content: contentLayer,
        selection: selectionLayer,
        debug: debugLayer,
        hitboxes: hitboxes,
    };
    return scene;
};
const isPointInBounds = (point, bounds) => {
    if (point.x > bounds.x &&
        point.x < bounds.x + bounds.width / 2 &&
        point.y > bounds.y &&
        point.y < bounds.y + bounds.height) {
        return 'left';
    }
    if (point.x > bounds.x + bounds.width / 2 &&
        point.x < bounds.x + bounds.width &&
        point.y > bounds.y &&
        point.y < bounds.y + bounds.height) {
        return 'right';
    }
    return undefined;
};
const getBounds = (child, translation) => {
    return child.type === 'group'
        ? {
            x: child.x + translation.x,
            y: child.y + translation.y - child.bounds.height,
            width: child.bounds.width,
            height: child.bounds.height + child.bounds.depth,
        }
        : {
            x: child.x + translation.x,
            y: child.y + translation.y,
            width: child.width,
            height: child.height,
        };
};
const getIntersection = (child, side) => {
    if (child.id) {
        return {
            type: 'content',
            id: child.id,
            side: side,
        };
    }
    else if (child.type === 'rect' && child.flag) {
        return {
            type: 'padding',
            flag: child.flag,
        };
    }
};
const findIntersections = (point, node, // must be the group containing the debug bounding rectangles
translation) => {
    const result = [];
    translation = translation || {
        x: node.x,
        y: node.y,
    };
    for (const child of node.children) {
        if (child.type !== 'group' && child.type !== 'rect') {
            throw new Error('Unexpected node type in hitboxes');
        }
        const bounds = getBounds(child, translation);
        const side = isPointInBounds(point, bounds);
        if (side) {
            const intersection = getIntersection(child, side);
            if (intersection) {
                result.push(intersection);
            }
            if (child.type === 'group') {
                const newTranslation = {
                    x: child.x + translation.x,
                    y: child.y + translation.y,
                };
                result.push(...findIntersections(point, child, newTranslation));
            }
        }
    }
    if (result.length === 0) {
        // iterate through all of the children again and find any bounds that
        // intersect with the vertical line crossing with `point` on it
        const candidateIntersections = [];
        for (const child of node.children) {
            if (child.type !== 'group' && child.type !== 'rect') {
                throw new Error('Unexpected node type in hitboxes');
            }
            const bounds = getBounds(child, translation);
            // We decide which side of the child to put the cursor on once we
            // decide on which candidate is the closest vertically to the point.
            if (point.x > bounds.x && point.x < bounds.x + bounds.width) {
                candidateIntersections.push(child);
            }
        }
        // If there are no candidates don't bother finding the closest one.
        if (candidateIntersections.length === 0) {
            return result;
        }
        const [int, ...rest] = candidateIntersections;
        // The closest candidate is the one with the smallest vertical distance
        // from the point.
        let bounds = getBounds(int, translation);
        let minDist = Math.min(Math.abs(bounds.y - point.y), Math.abs(bounds.y + bounds.height - point.y));
        let closest = int;
        for (const candidate of rest) {
            bounds = getBounds(candidate, translation);
            const dist = Math.min(Math.abs(bounds.y - point.y), Math.abs(bounds.y + bounds.height - point.y));
            if (dist < minDist) {
                minDist = dist;
                closest = candidate;
            }
        }
        // Determine whether the cursor should be to the left or the right of
        // the closest node.
        bounds = getBounds(closest, translation);
        const side = point.x < bounds.x + bounds.width / 2 ? 'left' : 'right';
        if (side) {
            const intersection = getIntersection(closest, side);
            if (intersection) {
                result.push(intersection);
            }
            if (closest.type === 'group') {
                const newTranslation = {
                    x: closest.x + translation.x,
                    y: closest.y + translation.y,
                };
                result.push(...findIntersections(point, closest, newTranslation));
            }
        }
    }
    return result;
};

var sceneGraph = /*#__PURE__*/Object.freeze({
    __proto__: null,
    findIntersections: findIntersections,
    processBox: processBox
});

const typesetDelimited = (typesetChild, node, context) => {
    const thresholdOptions = {
        value: 'both',
        strict: true,
    };
    const row = typesetChild(0, context);
    if (!row) {
        throw new Error("Delimited's content should be defined");
    }
    const open = makeDelimiter(node.leftDelim.value, row, thresholdOptions, context);
    const close = makeDelimiter(node.rightDelim.value, row, thresholdOptions, context);
    open.pending = node.leftDelim.pending;
    close.pending = node.rightDelim.pending;
    const delimited = makeStaticHBox([open, row, close], context);
    delimited.id = node.id;
    delimited.style = node.style;
    return delimited;
};

const typesetMacro = (typesetChild, node, context) => {
    const row = typesetChild(0, { ...context, macro: true });
    if (!row) {
        throw new Error("Delimited's content should be defined");
    }
    const { font } = context.fontData;
    const glyphID = font.getGlyphID('\\');
    const slash = makeGlyph('\\', glyphID, context, false);
    // TODO: add an outline rect that matches the size of the box
    const macro = makeStaticHBox([slash, row], context);
    macro.id = node.id;
    macro.style = node.style;
    return macro;
};

const makeList = (size, box) => [
    makeKern(size),
    box,
];
const childContextForFrac = (context) => {
    const { mathStyle } = context;
    const childMathStyle = {
        [exports.MathStyle.Display]: exports.MathStyle.Text,
        [exports.MathStyle.Text]: exports.MathStyle.Script,
        [exports.MathStyle.Script]: exports.MathStyle.ScriptScript,
        [exports.MathStyle.ScriptScript]: exports.MathStyle.ScriptScript,
    }[mathStyle];
    const childContext = {
        ...context,
        mathStyle: childMathStyle,
    };
    return childContext;
};
const typesetFrac = (typesetChild, node, context) => {
    const childContext = childContextForFrac(context);
    let numBox = typesetChild(0, childContext);
    let denBox = typesetChild(1, childContext);
    if (!numBox || !denBox) {
        throw new Error('The numerator and denominator must both be defined');
    }
    const { mathStyle } = context;
    const { constants } = context.fontData.font.math;
    const fontSize = fontSizeForContext(context);
    const thickness = fontSize * constants.fractionRuleThickness;
    const shift = fontSize * constants.axisHeight;
    // If useDisplayStyle is false then we need to reduce the font size of
    // numerators and denominators
    const useDisplayStyle = mathStyle === exports.MathStyle.Display;
    const minDenGap = useDisplayStyle
        ? fontSize * constants.fractionDenomDisplayStyleGapMin
        : fontSize * constants.fractionDenominatorGapMin;
    const minNumGap = useDisplayStyle
        ? fontSize * constants.fractionNumDisplayStyleGapMin
        : fontSize * constants.fractionNumeratorGapMin;
    const multiplier = multiplierForContext(context);
    const endPadding = thickness; // add extra space around the numerator and denominator
    const width = Math.max(Math.max(getWidth(numBox), getWidth(denBox)), fontSize * multiplier * 0.5) +
        2 * endPadding;
    // center the numerator
    if (getWidth(numBox) < width) {
        const kernSize = (width - getWidth(numBox)) / 2;
        numBox = rebox(numBox, makeKern(kernSize, 'start'), makeKern(kernSize, 'end'));
    }
    // center the denominator
    if (getWidth(denBox) < width) {
        const kernSize = (width - getWidth(denBox)) / 2;
        denBox = rebox(denBox, makeKern(kernSize, 'start'), makeKern(kernSize, 'end'));
    }
    const upList = makeList(minNumGap, numBox);
    const dnList = makeList(minDenGap, denBox);
    const stroke = makeStaticHBox([makeHRule(thickness, width)], context);
    const fracBox = makeVBox(width, stroke, upList, dnList, context);
    fracBox.shift = -shift;
    fracBox.id = node.id;
    fracBox.style = node.style;
    return fracBox;
};

const childContextForLimits = (context) => {
    const { mathStyle } = context;
    const childMathStyle = {
        [exports.MathStyle.Display]: exports.MathStyle.Script,
        [exports.MathStyle.Text]: exports.MathStyle.Script,
        [exports.MathStyle.Script]: exports.MathStyle.ScriptScript,
        [exports.MathStyle.ScriptScript]: exports.MathStyle.ScriptScript,
    }[mathStyle];
    const childContext = {
        ...context,
        mathStyle: childMathStyle,
        cramped: true,
    };
    return childContext;
};
// TODO: render as a subsup if context.mathStyle isn't MathStyle.Display
const typesetLimits = (typesetChild, node, path, context, typesetNode) => {
    const childContext = childContextForLimits(context);
    const lowerBox = typesetChild(0, childContext);
    const upperBox = typesetChild(1, childContext);
    if (!lowerBox) {
        throw new Error('Lower limit should always be defined');
    }
    // TODO: figure out what the path should be for `inner`
    const inner = typesetNode(node.inner, path, {
        ...context,
        operator: true,
    });
    inner.id = node.inner.id;
    inner.style = {
        ...inner.style,
        color: node.inner.style.color,
    };
    const innerWidth = getWidth(inner);
    const width = Math.max(innerWidth, lowerBox.width || 0, (upperBox === null || upperBox === void 0 ? void 0 : upperBox.width) || 0);
    const newInner = innerWidth < width
        ? makeStaticHBox([
            makeKern((width - innerWidth) / 2),
            inner,
            makeKern((width - innerWidth) / 2),
        ], context)
        : inner;
    if (lowerBox.width < width) {
        lowerBox.shift = (width - lowerBox.width) / 2;
    }
    if (upperBox && upperBox.width < width) {
        upperBox.shift = (width - upperBox.width) / 2;
    }
    const limits = makeVBox(width, newInner, upperBox ? [makeKern(6), upperBox] : [], [makeKern(4), lowerBox], context);
    limits.id = node.id;
    limits.style = node.style;
    return limits;
};

const typesetRoot = (typesetChild, node, context) => {
    var _a;
    const degreeContext = {
        ...context,
        // It doesn't matter what the mathStyle is of the parent, we
        // always use ScriptScript for root indicies.
        mathStyle: exports.MathStyle.ScriptScript,
    };
    const degree = typesetChild(0, degreeContext);
    const radicand = typesetChild(1, context);
    if (!radicand) {
        throw new Error('Radicand must be defined');
    }
    const multiplier = multiplierForContext(context);
    // Give the radicand a minimal width in case it's empty
    radicand.width = Math.max(radicand.width, 30 * multiplier);
    const thresholdOptions = {
        value: 'sum',
        strict: true,
    };
    const surd = makeStaticHBox([makeDelimiter('\u221A', radicand, thresholdOptions, context)], context);
    const fontSize = fontSizeForContext(context);
    const { constants } = context.fontData.font.math;
    const thickness = fontSize * constants.radicalRuleThickness;
    const endPadding = thickness; // Add extra space at the end of the radicand
    const stroke = makeHRule(thickness, radicand.width + endPadding);
    const radicalWithRule = makeVBox(radicand.width, radicand, [makeKern(6), stroke], [], context);
    // Compute the shift to align the top of the surd with the radical rule
    const shift = surd.height - radicalWithRule.height;
    surd.shift = shift;
    let root;
    if (degree) {
        const afterDegreeKern = makeKern(fontSize * constants.radicalKernAfterDegree);
        // TODO: take into account constants.radicalKernBeforeDegree
        const beforeDegreeKern = makeKern(Math.max(0, -afterDegreeKern.size - degree.width));
        // This follows the instructions from 3.3.3.3 describing the alphabetic
        // baseline of the index.
        // https://mathml-refresh.github.io/mathml-core/#root-with-index
        //
        // NOTE: This doesn't account for an index with a large descender, but
        // neither does TeX's layout algorithm.  Most of the the time the index
        // is a single number or `n`, neither of whcih have a descender.
        const degreeBottomRaisePercent = constants.radicalDegreeBottomRaisePercent / 100;
        // We default to MathML/Word beahavior since that's what most fonts
        // seem to use.
        const algorithm = (_a = context.radicalDegreeAlgorithm) !== null && _a !== void 0 ? _a : exports.RadicalDegreeAlgorithm.MathML;
        switch (algorithm) {
            case exports.RadicalDegreeAlgorithm.OpenType:
                degree.shift =
                    shift - // match shift of surdHBox
                        // The OpenType spec says `radicalDegreeBottomRaisePercent` is
                        // with respect to the ascender of the radical glyph.
                        degreeBottomRaisePercent * surd.height;
                break;
            case exports.RadicalDegreeAlgorithm.MathML:
                degree.shift =
                    shift + // match shift of surdHBox
                        surd.depth - // align the index to the bottom of surdHBox
                        // The MathML Core spec says `radicalDegreeBottomRaisePercent`
                        // is with respect to the height of the radical glyph.
                        degreeBottomRaisePercent * vsize(surd);
                break;
        }
        root = makeStaticHBox([beforeDegreeKern, degree, afterDegreeKern, surd, radicalWithRule], context);
    }
    else {
        root = makeStaticHBox([surd, radicalWithRule], context);
    }
    root.width += endPadding;
    root.id = node.id;
    root.style = node.style;
    return root;
};

const childContextForSubsup = (context) => {
    const { mathStyle } = context;
    const childMathStyle = {
        [exports.MathStyle.Display]: exports.MathStyle.Script,
        [exports.MathStyle.Text]: exports.MathStyle.Script,
        [exports.MathStyle.Script]: exports.MathStyle.ScriptScript,
        [exports.MathStyle.ScriptScript]: exports.MathStyle.ScriptScript,
    }[mathStyle];
    const childContext = {
        ...context,
        mathStyle: childMathStyle,
        cramped: true,
    };
    return childContext;
};
const typesetSubsup = (typesetChild, node, context, prevEditNode, prevLayoutNode) => {
    const childContext = childContextForSubsup(context);
    const subBox = typesetChild(0, childContext);
    const supBox = typesetChild(1, childContext);
    if (!supBox && !subBox) {
        throw new Error('at least one of supBox and subBox must be defined');
    }
    const { font } = context.fontData;
    const width = Math.max(supBox ? getWidth(supBox) : 0, subBox ? getWidth(subBox) : 0);
    // Some atoms gets wrapped in a box to add padding to them so we need to
    // filter them out.  Anything else that's in a box is some sort of compound
    // layout structure (frac, delimited, etc.) and should have its subscript
    // and/or superscript positioned based on the size of the box.
    if ((prevEditNode === null || prevEditNode === void 0 ? void 0 : prevEditNode.type) !== 'char' && prevLayoutNode) {
        const superscriptBaselineDropMax = getConstantValue(font.math.constants.superscriptBaselineDropMax, context);
        const subscriptBaselineDropMin = getConstantValue(font.math.constants.subscriptBaselineDropMin, context);
        const upList = [];
        const dnList = [];
        if (supBox) {
            const shift = getHeight(prevLayoutNode);
            const kernShift = shift - superscriptBaselineDropMax;
            upList.push(makeKern(kernShift));
            upList.push(supBox);
        }
        if (subBox) {
            const shift = getDepth(prevLayoutNode);
            const kernSize = shift - subscriptBaselineDropMin;
            dnList.push(makeKern(kernSize));
            dnList.push(subBox);
        }
        const referenceNode = makeKern(0); // empty reference node
        const subsupBox = makeVBox(width, referenceNode, upList, dnList, context);
        subsupBox.id = node.id;
        subsupBox.style = node.style;
        return subsupBox;
    }
    const upList = [];
    const fontSize = fontSizeForContext(context);
    const childFontSize = fontSizeForContext(childContext);
    const parenMetrics = font.getGlyphMetrics(font.getGlyphID(')'));
    const overshoot = (font.head.unitsPerEm - parenMetrics.height) / 2;
    if (supBox) {
        const xMetrics = font.getGlyphMetrics(font.getGlyphID('x'));
        const xHeight = (xMetrics.bearingY * fontSize) / font.head.unitsPerEm;
        const maxDepth = ((parenMetrics.height - parenMetrics.bearingY + overshoot) *
            childFontSize) /
            font.head.unitsPerEm;
        // Compute shift in baseline of superscript.
        // The baseline is at the same height as the xHeight of the parent.
        let kernSize = -supBox.depth + xHeight;
        if (supBox.depth > maxDepth) {
            // If the superscript's depth is greater that of a row without any non-atom
            // childre, then push the superscript up by whatever the difference is.
            kernSize += supBox.depth - maxDepth;
        }
        upList.push(makeKern(kernSize));
        upList.push(supBox);
    }
    const dnList = [];
    if (subBox) {
        const xMetrics = font.getGlyphMetrics(font.getGlyphID('x'));
        const xHeight = (xMetrics.bearingY * childFontSize) / font.head.unitsPerEm;
        const maxHeight = ((parenMetrics.bearingY + overshoot) * childFontSize) /
            font.head.unitsPerEm;
        // Compute shift in baseline of subscript.
        // The x-height of the subscript is the same as baseline of the character to
        // the left of the subsup.
        let kernSize = -subBox.height + xHeight;
        if (subBox.height > maxHeight) {
            // If the subscript's height is greater that of a row without any non-atom
            // childre, then push the subscript down by whatever the difference is.
            kernSize += subBox.height - maxHeight;
        }
        dnList.push(makeKern(kernSize));
        dnList.push(subBox);
    }
    const referenceNode = makeKern(0); // empty reference node
    const subsupBox = makeVBox(width, referenceNode, upList, dnList, context);
    subsupBox.id = node.id;
    subsupBox.style = node.style;
    return subsupBox;
};

const DEFAULT_GUTTER_WIDTH = 50;
const isOperator = (cell) => (cell === null || cell === void 0 ? void 0 : cell.children.length) === 1 &&
    Editor__namespace.util.isAtom(cell.children[0], ['+', '\u2212', '=', '<', '>']);
const childContextForTable = (context) => {
    const { mathStyle } = context;
    const childMathStyle = {
        [exports.MathStyle.Display]: exports.MathStyle.Text,
        [exports.MathStyle.Text]: exports.MathStyle.Script,
        [exports.MathStyle.Script]: exports.MathStyle.ScriptScript,
        [exports.MathStyle.ScriptScript]: exports.MathStyle.ScriptScript,
    }[mathStyle];
    const childContext = {
        ...context,
        mathStyle: childMathStyle,
    };
    return childContext;
};
const typesetTable = (typesetChild, node, context, zipper) => {
    var _a;
    const columns = [];
    const rows = [];
    const childContext = childContextForTable(context);
    const reboxColumn = (col, leftKernSize, rightKernSize) => {
        for (let row = 0; row < node.rowCount; row++) {
            let cell = rows[row].children[col];
            cell = rebox(cell, makeKern(leftKernSize), makeKern(rightKernSize));
            rows[row].children[col] = cell;
            columns[col].children[row] = cell;
            columns[col].width = Math.max(columns[col].width, cell.width);
        }
    };
    const gutterWidth = node.subtype === 'algebra' ? 0 : DEFAULT_GUTTER_WIDTH;
    const children = node.type === 'table'
        ? node.children
        : [
            ...node.left,
            // @ts-expect-error: zipper is always defined when
            // node is a ZTable
            Editor__namespace.zrowToRow(zipper.row),
            ...node.right,
        ];
    const MIN_WIDTH = 0; // 32; // only use this for debugging purposes.
    // Group cells into rows and columns and determine the width of each
    // column and the depth/height of each row.
    for (let j = 0; j < node.rowCount; j++) {
        for (let i = 0; i < node.colCount; i++) {
            if (!columns[i]) {
                columns[i] = {
                    children: [],
                    width: MIN_WIDTH,
                };
            }
            if (!rows[j]) {
                rows[j] = {
                    children: [],
                    height: 0,
                    depth: 0,
                };
            }
            // We want to add padding around the first operator if it's the only
            // character in the cell.
            const child = children[j * node.colCount + i];
            const padFirstOperator = ((_a = child === null || child === void 0 ? void 0 : child.children) === null || _a === void 0 ? void 0 : _a.length) === 1 &&
                child.children[0].type === 'char' &&
                ['+', '\u2212'].includes(child.children[0].value);
            let cell = typesetChild(j * node.colCount + i, childContext, padFirstOperator);
            if (cell) {
                columns[i].width = Math.max(cell.width, columns[i].width);
                rows[j].height = Math.max(cell.height, rows[j].height);
                rows[j].depth = Math.max(cell.depth, rows[j].depth);
            }
            else {
                // Use an empty Layout.Box for children that were null
                cell = {
                    type: 'HBox',
                    shift: 0,
                    content: {
                        type: 'static',
                        nodes: [],
                    },
                    // These values don't matter since the box is empty
                    fontSize: 0,
                    style: {},
                    // These will get filled in later
                    width: 0,
                    height: 0,
                    depth: 0,
                };
            }
            columns[i].children.push(cell);
            rows[j].children.push(cell);
        }
    }
    const cursorIndex = node.type === 'ztable' ? node.left.length : -1;
    if (node.subtype === 'algebra' && cursorIndex !== -1) {
        const col = cursorIndex % node.colCount;
        const zrow = zipper === null || zipper === void 0 ? void 0 : zipper.row;
        const cellColumns = [];
        for (let i = 0; i < node.colCount; i++) {
            const col = [];
            for (let j = 0; j < node.rowCount; j++) {
                const index = j * node.colCount + i;
                // TODO: check that children[index] isn't null
                col.push(children[index]);
            }
            cellColumns.push(col); // this is unsafe
        }
        if (zrow) {
            if (col === 0 && !cellColumns[col + 1].some(isOperator)) {
                // Add right padding on every cell in the first column
                reboxColumn(col, 0, 16);
            }
            else if (col === node.colCount - 1 &&
                !cellColumns[col - 1].some(isOperator)) {
                // Add left padding on every cell in the last column
                reboxColumn(col, 16, 0);
            }
            else if (col > 0 &&
                col < cellColumns.length - 1 &&
                Editor__namespace.isColumnEmpty(cellColumns[col])) {
                // If the cursor is in an empty column, only add padding to one
                // side of the column if there's an operator in one of the
                // columns. The padding goes on the opposite side of the column
                // with the operator since operators have their own built-in
                // padding.
                if (cellColumns[col - 1].some(isOperator) &&
                    !cellColumns[col + 1].some(isOperator)) {
                    reboxColumn(col, 0, 16);
                }
                else if (cellColumns[col + 1].some(isOperator) &&
                    !cellColumns[col - 1].some(isOperator)) {
                    reboxColumn(col, 16, 0);
                }
            }
        }
        // If there are any columns with no operators on either side, add both
        // left and right padding regardless of whether the cursor is in the
        // column or not.
        for (let col = 0; col < node.colCount; col++) {
            if (col > 0 &&
                col < cellColumns.length - 1 &&
                Editor__namespace.isColumnEmpty(cellColumns[col]) &&
                !cellColumns[col - 1].some(isOperator) &&
                !cellColumns[col + 1].some(isOperator)) {
                reboxColumn(col, 16, 16);
            }
        }
    }
    // Adjust the width of cells in the same column to be the same
    for (let i = 0; i < columns.length; i++) {
        const col = columns[i];
        for (let j = 0; j < col.children.length; j++) {
            // center the cell content
            const originalWidth = getWidth(col.children[j]);
            const baseKernSize = (col.width - originalWidth) / 2;
            let rightKernSize = i < columns.length - 1 ? baseKernSize + gutterWidth / 2 : baseKernSize;
            let leftKernSize = i > 0 ? baseKernSize + gutterWidth / 2 : baseKernSize;
            const child = children[j * node.colCount + i];
            // Right align any child that has content when showing work
            if (node.subtype === 'algebra' && child && child.children.length > 0) {
                leftKernSize = baseKernSize * 2;
                rightKernSize = 0;
            }
            const cell = rebox(col.children[j], makeKern(leftKernSize, 'start'), makeKern(rightKernSize, 'end'));
            col.children[j] = cell;
            // rows[] has its own references so we need to update it as well
            rows[j].children[i] = cell;
        }
    }
    // Adjust the height/depth of cells in the same row to be the same
    for (let i = 0; i < rows.length; i++) {
        const row = rows[i];
        for (let j = 0; j < row.children.length; j++) {
            row.children[j].height = row.height;
            row.children[j].depth = row.depth;
        }
    }
    const width = columns.reduce((sum, col) => sum + col.width, 0) +
        gutterWidth * (columns.length - 1);
    const { constants } = context.fontData.font.math;
    const fontSize = fontSizeForContext(context);
    const rowBoxes = rows.map((row, index) => {
        var _a;
        const result = makeStaticHBox(row.children, context);
        const style = (_a = node.rowStyles) === null || _a === void 0 ? void 0 : _a[index];
        if ((style === null || style === void 0 ? void 0 : style.border) === 'top') {
            const thickness = fontSize * constants.fractionRuleThickness;
            const stroke = makeStaticHBox([makeHRule(thickness, width)], context);
            const { mathStyle } = context;
            const useDisplayStyle = mathStyle === exports.MathStyle.Display;
            const minDenGap = useDisplayStyle
                ? fontSize * constants.fractionDenomDisplayStyleGapMin
                : fontSize * constants.fractionDenominatorGapMin;
            const minNumGap = useDisplayStyle
                ? fontSize * constants.fractionNumDisplayStyleGapMin
                : fontSize * constants.fractionNumeratorGapMin;
            return makeVBox(width, stroke, [makeKern(minNumGap)], [makeKern(minDenGap), result], context);
        }
        return result;
    });
    const inner = makeVBox(width, rowBoxes[0], [], rowBoxes.slice(1), context);
    const shift = fontSize * constants.axisHeight;
    // Equalize the depth and height and then shift up so the center of the
    // table aligns with the central axis.
    const vsize$1 = vsize(inner);
    inner.height = vsize$1 / 2 + shift;
    inner.depth = vsize$1 / 2 - shift;
    const thresholdOptions = {
        value: 'sum',
        strict: true,
    };
    if (!node.delimiters) {
        inner.id = node.id;
        inner.style = node.style;
        return inner;
    }
    const open = makeDelimiter(node.delimiters.left.value, inner, thresholdOptions, context);
    const close = makeDelimiter(node.delimiters.right.value, inner, thresholdOptions, context);
    const table = makeStaticHBox([open, inner, close], context);
    table.id = node.id;
    table.style = node.style;
    return table;
};

const canBeUnary = (char) => {
    const unaryOperators = [
        '+',
        '\u2212', // \minus
        '\u00B1', // \pm
    ];
    return unaryOperators.includes(char);
};
const shouldHavePadding = (prevNode, currentNode, context) => {
    const currentChar = currentNode.value;
    // We only add padding around operators, so if we get a non-operator char
    // we can return early.
    if (!Editor__namespace.util.isOperator(currentNode)) {
        return false;
    }
    // No operators get padding when `cramped` is true
    if (context.cramped) {
        return false;
    }
    // If the currentChar can be unary we check a number of situations where it
    // should be unary and don't give it any padding in those situations.
    if (canBeUnary(currentChar)) {
        if (!prevNode ||
            (prevNode.type === 'char' && Editor__namespace.util.isOperator(prevNode)) ||
            prevNode.type === 'limits' ||
            prevNode.type === 'zlimits') {
            return false;
        }
    }
    // All other operators should have padding around them.
    return true;
};
const maybeAddOperatorPadding = (prevNode, currentNode, context, padOperator) => {
    const glyph = typesetAtom(currentNode, context);
    const fontSize = fontSizeForContext(context);
    const result = (padOperator && Editor__namespace.util.isOperator(currentNode)) ||
        shouldHavePadding(prevNode, currentNode, context)
        ? makeStaticHBox([makeKern(fontSize / 4), glyph, makeKern(fontSize / 4)], context)
        : glyph;
    if (result !== glyph) {
        result.id = glyph.id;
        delete glyph.id;
        // Move the style to the result so that cancel overlays are
        // continuous even when they include an operator with padding.
        result.style = glyph.style;
        glyph.style = {};
    }
    return result;
};
const typesetAtom = (node, context) => {
    const { font } = context.fontData;
    const glyphID = font.getGlyphID(node.value);
    let glyph = makeGlyph(node.value, glyphID, context);
    // Convert individual glyphs to italic glyphs if they exist in the
    // current font.
    if (/[a-z]/.test(node.value) && !context.operator) {
        const offset = node.value.charCodeAt(0) - 'a'.charCodeAt(0);
        const char = context.macro
            ? node.value
            : String.fromCodePoint(0x1d44e + offset);
        const glyphID = font.getGlyphID(char);
        glyph = makeGlyph(char, glyphID, context);
    }
    glyph.id = node.id;
    glyph.style = node.style;
    glyph.pending = node.pending;
    return glyph;
};

const { NodeType, SelectionUtils, PathUtils } = Editor__namespace;
const typesetRow = (row, path, context, padFirstOperator) => {
    const output = typesetNodes(row.children, path, context, undefined, undefined, padFirstOperator);
    const { selection } = context;
    if (selection) {
        const { path: selectionPath, start, end, } = SelectionUtils.getPathAndRange(selection);
        if (PathUtils.equals(path, selectionPath)) {
            const left = output.slice(0, start);
            const middle = output.slice(start, end);
            const right = output.slice(end);
            const box = (SelectionUtils.isCollapsed(selection)
                ? makeCursorHBox(left, right, context)
                : makeSelectionHBox(left, middle, right, context));
            box.id = row.id;
            box.style = {
                ...box.style,
                color: row.style.color,
            };
            if (context.renderMode === exports.RenderMode.Dynamic) {
                ensureMinDepthAndHeight(box, context);
            }
            return box;
        }
    }
    const box = makeStaticHBox(output, context);
    box.id = row.id;
    box.style = {
        ...box.style,
        color: row.style.color,
    };
    if (context.renderMode === exports.RenderMode.Dynamic) {
        ensureMinDepthAndHeight(box, context);
    }
    return box;
};
/**
 * This function is used to guarantee a certain depth/height for a box that is
 * equal to the depth/height of the cursor or selection rectangle.  This is
 * useful for typesetting while editing since it minimize changes to vertical
 * size of the rendered math.
 *
 * WARNING: This function mutates `box`.
 *
 * TODO: add originalDepth and originalHeight so that getDelimiter can make its
 * decisions based on the original dimensions of the box.
 *
 * @param {Mutable<Dim>} dim
 * @param {Context} context
 * @return {void}
 */
const ensureMinDepthAndHeight = (dim, context) => {
    const { fontData: { font }, } = context;
    const fontSize = fontSizeForContext(context);
    const parenMetrics = font.getGlyphMetrics(font.getGlyphID(')'));
    // This assumes that parenMetrics.height < font.head.unitsPerEm
    const overshoot = (font.head.unitsPerEm - parenMetrics.height) / 2;
    const depth = ((parenMetrics.height - parenMetrics.bearingY + overshoot) * fontSize) /
        font.head.unitsPerEm;
    dim.depth = Math.max(dim.depth, depth);
    const height = ((parenMetrics.bearingY + overshoot) * fontSize) / font.head.unitsPerEm;
    dim.height = Math.max(dim.height, height);
};
const getTypesetChildFromZipper = (zipper, focus, path) => {
    return (index, context, padFirstOperator) => {
        if (index < focus.left.length) {
            const child = focus.left[index];
            return (child && typesetRow(child, [...path, index], context, padFirstOperator));
        }
        else if (index === focus.left.length) {
            return _typesetZipper(zipper, path, context, padFirstOperator);
        }
        else {
            const child = focus.right[index - focus.left.length - 1];
            return (child && typesetRow(child, [...path, index], context, padFirstOperator));
        }
    };
};
const getTypesetChildFromNodes = (children, path) => {
    return (index, context, padFirstOperator) => {
        const child = children[index];
        return (child && typesetRow(child, [...path, index], context, padFirstOperator));
    };
};
const typesetFocus = (focus, zipper, path, context, prevEditNode, prevLayoutNode) => {
    const typesetChild = getTypesetChildFromZipper(zipper, focus, path);
    switch (focus.type) {
        case 'zfrac': {
            return typesetFrac(typesetChild, focus, context);
        }
        case 'zsubsup': {
            return typesetSubsup(typesetChild, focus, context, prevEditNode, prevLayoutNode);
        }
        case 'zroot': {
            return typesetRoot(typesetChild, focus, context);
        }
        case 'zlimits': {
            return typesetLimits(typesetChild, focus, path, context, typesetNode);
        }
        case 'zdelimited': {
            return typesetDelimited(typesetChild, focus, context);
        }
        case 'ztable': {
            return typesetTable(typesetChild, focus, context, zipper);
        }
        default:
            throw new core.UnreachableCaseError(focus);
    }
};
const typesetNode = (node, path, context, prevEditNode, prevLayoutNode, padFirstOperator) => {
    switch (node.type) {
        case NodeType.Row: {
            // The only time this can happen is if limits.inner is a row
            return typesetRow(node, path, context);
        }
        case NodeType.Frac: {
            const typesetChild = getTypesetChildFromNodes(node.children, path);
            return typesetFrac(typesetChild, node, context);
        }
        case NodeType.SubSup: {
            return typesetSubsup(getTypesetChildFromNodes(node.children, path), node, context, prevEditNode, prevLayoutNode);
        }
        case NodeType.Root: {
            const typesetChild = getTypesetChildFromNodes(node.children, path);
            return typesetRoot(typesetChild, node, context);
        }
        case NodeType.Limits: {
            const typesetChild = getTypesetChildFromNodes(node.children, path);
            return typesetLimits(typesetChild, node, path, context, typesetNode);
        }
        case NodeType.Delimited: {
            const typesetChild = getTypesetChildFromNodes(node.children, path);
            return typesetDelimited(typesetChild, node, context);
        }
        case NodeType.Macro: {
            const typesetChild = getTypesetChildFromNodes(node.children, path);
            return typesetMacro(typesetChild, node, context);
        }
        case NodeType.Table: {
            const typesetChild = getTypesetChildFromNodes(node.children, path);
            return typesetTable(typesetChild, node, context);
        }
        case 'char': {
            return maybeAddOperatorPadding(prevEditNode, node, context, padFirstOperator);
        }
        default:
            throw new core.UnreachableCaseError(node);
    }
};
const typesetNodes = (nodes, path, context, prevChild, prevLayoutNode, padFirstOperator) => {
    return nodes.map((child, index) => {
        const result = typesetNode(child, [...path, index], context, prevChild, prevLayoutNode, index === 0 ? padFirstOperator : undefined);
        prevLayoutNode = result;
        prevChild = child;
        return result;
    });
};
const _typesetZipper = (zipper, path, context, padFirstOperator) => {
    // The bottommost crumb is the outermost row
    const [crumb, ...restCrumbs] = zipper.breadcrumbs;
    if (crumb) {
        const row = crumb.row;
        const nextZipper = {
            ...zipper,
            breadcrumbs: restCrumbs,
        };
        const nodes = [];
        nodes.push(...typesetNodes(row.left, path, context, undefined, undefined, padFirstOperator));
        nodes.push(typesetFocus(crumb.focus, nextZipper, path, context, row.left[row.left.length - 1], // previous edit node
        nodes[nodes.length - 1]));
        nodes.push(...typesetNodes(row.right, path, context, crumb.focus, // previous edit node
        nodes[nodes.length - 1], // previous layout node
        padFirstOperator));
        const box = makeStaticHBox(nodes, context);
        box.id = row.id;
        box.style = {
            ...box.style,
            color: row.style.color,
        };
        if (context.renderMode === exports.RenderMode.Dynamic) {
            ensureMinDepthAndHeight(box, context);
        }
        return box;
    }
    else {
        const row = zipper.row;
        const input = [...row.left, ...row.selection, ...row.right];
        const output = typesetNodes(input, path, context, undefined, undefined, padFirstOperator);
        const firstCut = row.left.length;
        const secondCut = firstCut + row.selection.length;
        const left = output.slice(0, firstCut);
        const selection = output.slice(firstCut, secondCut);
        const right = output.slice(secondCut);
        const box = (selection.length > 0
            ? makeSelectionHBox(left, selection, right, context)
            : makeCursorHBox(left, right, context));
        box.id = row.id;
        box.style = {
            ...box.style,
            color: row.style.color,
        };
        if (context.renderMode === exports.RenderMode.Dynamic) {
            ensureMinDepthAndHeight(box, context);
        }
        return box;
    }
};
const typesetZipper = (zipper, context, options = {}) => {
    const path = [];
    const box = _typesetZipper(zipper, path, context);
    return processBox(box, context.fontData, options);
};
const typeset = (node, context, options = {}) => {
    const path = [];
    const box = typesetNode(node, path, context);
    return processBox(box, context.fontData, options);
};

exports.SceneGraph = sceneGraph;
exports.typeset = typeset;
exports.typesetZipper = typesetZipper;
